@@ -83,6 +83,8 @@ struct EBookBackendExchangePrivate {
 	gboolean is_cache_ready;
 	gboolean marked_for_offline;
 
+	GMutex *create_mutex;
+
 	EBookBackendSummary *summary;
 	EBookBackendCache *cache;
 };
@@ -95,14 +97,14 @@ static void proppatch_email(PropMapping *prop_mapping, EContact *new_contact, EC
 static void proppatch_date(PropMapping *prop_mapping, EContact *new_contact, EContact *cur_contact, E2kProperties *props);
 static void populate_address(EContactField field, EContact *new_contact, void *data);
 static void populate_date(EContactField field, EContact *new_contact, void *data);
-#ifdef ENABLE_CATEGORIES
-static void proppatch_categories(PropMapping *prop_mapping, EContact *new_contact, EContact *cur_contact, E2kProperties *props);
-static void populate_categories(EContactField field, EContact *new_contact, void *data);
-#endif
+static void proppatch_categories (PropMapping *prop_mapping, EContact *new_contact, EContact *cur_contact, E2kProperties *props);
+static void populate_categories (EContactField field, EContact *new_contact, void *data);
 static E2kRestriction *e_book_backend_exchange_build_restriction (const char *sexp,
 							       E2kRestriction *base_rn);
 
 static const char *contact_name (EContact *contact);
+static void proppatch_im (PropMapping *prop_mapping, EContact *new_contact, EContact *cur_contact, E2kProperties *props);
+static void populate_im (EContactField field, EContact *new_contact, void *data);
 
 static GPtrArray *field_names_array;
 static const char **field_names;
@@ -133,6 +135,7 @@ http_status_to_pas (E2kHTTPStatus status)
 #define CAL_FIELD(x,z) { E_CONTACT_##x, E2K_PR_CALENDAR_##x, z, FLAG_UNLIKEABLE, NULL }
 #define HTTPMAIL_FIELD(x,y,z) { E_CONTACT_##x, E2K_PR_HTTPMAIL_##y, z, FLAG_UNLIKEABLE | FLAG_PUT, NULL }
 #define EXCHANGE_FIELD(x,y,z,w,v) { E_CONTACT_##x, E2K_PR_EXCHANGE_##y, z, FLAG_COMPOSITE, w, v }
+#define IM_FIELD(x,z) { E_CONTACT_##x, E2K_PR_OUTLOOK_CONTACT_IM_ADDR, z, FLAG_COMPOSITE, proppatch_im, populate_im }
 
 struct PropMapping {
 	EContactField field;
@@ -200,13 +203,16 @@ prop_mappings[] = {
 	DATE_FIELD (BIRTH_DATE, "birth_date"),
 	DATE_FIELD (ANNIVERSARY, "anniversary"),
 	CAL_FIELD (FREEBUSY_URL, "fburl"),
+	IM_FIELD (IM_ICQ, "icq"),
+	IM_FIELD (IM_AIM, "aim"),
+	IM_FIELD (IM_MSN, "msn"),
+	IM_FIELD (IM_YAHOO, "yahoo"),
+	IM_FIELD (IM_JABBER, "jabber"),
+	IM_FIELD (IM_GROUPWISE, "nov"),
 
 	HTTPMAIL_FIELD (NOTE, TEXT_DESCRIPTION, "note"),
 
-#ifdef ENABLE_CATEGORIES
-	/* this doesn't work at the moment */
 	EXCHANGE_FIELD (CATEGORIES, KEYWORDS, "categories", proppatch_categories, populate_categories)
-#endif
 };
 static int num_prop_mappings = sizeof(prop_mappings) / sizeof (prop_mappings[0]);
 
@@ -468,6 +474,7 @@ build_cache (EBookBackendExchange *be)
 				       field_names, n_field_names,
 				       bepriv->base_rn, NULL, TRUE);
 
+	e_file_cache_freeze_changes (E_FILE_CACHE (bepriv->cache));
 	while ((result = e2k_result_iter_next (iter))) {
 		contact = e_contact_from_props (be, result);
 		if (!contact)
@@ -477,6 +484,8 @@ build_cache (EBookBackendExchange *be)
 	}
 	e_book_backend_cache_set_populated (bepriv->cache);
 	bepriv->is_cache_ready=TRUE;
+	e_file_cache_thaw_changes (E_FILE_CACHE (bepriv->cache));
+
 	return NULL;
 }
 
@@ -509,6 +518,7 @@ update_cache (EBookBackendExchange *be)
 				       field_names, n_field_names,
 				       bepriv->base_rn, NULL, TRUE);
 
+	e_file_cache_freeze_changes (E_FILE_CACHE (bepriv->cache));
 	while ((result = e2k_result_iter_next (iter))) {
 		contact = e_contact_from_props (be, result);
 		if (!contact)
@@ -518,6 +528,8 @@ update_cache (EBookBackendExchange *be)
 	}
 	e_book_backend_cache_set_populated (bepriv->cache);
 	bepriv->is_cache_ready=TRUE;
+	e_file_cache_thaw_changes (E_FILE_CACHE (bepriv->cache));
+
 	return TRUE;
 }
 
@@ -748,82 +760,46 @@ proppatch_email (PropMapping *prop_mapping,
 		e2k_properties_remove (props, E2K_PR_MAPI_EMAIL_ADDRESS_LIST);
 }
 
-#ifdef ENABLE_CATEGORIES
 static void
 proppatch_categories (PropMapping *prop_mapping,
 		      EContact *new_contact, EContact *cur_contact,
 		      E2kProperties *props)
 {
-	gboolean changed;
-	char *new_categories, *cur_categories = NULL;
-	EList *categories;
-	EIterator *iter;
-	GPtrArray *array = NULL;
+	GList *categories_list = NULL;
+	GList *l = NULL;
+	GPtrArray *prop_array = NULL;
 
-	new_categories = e_contact_get (new_contact, field);
-	if (cur_contact)
-		cur_categories = e_contact_get (cur_contact, field);
+	categories_list = e_contact_get (new_contact, E_CONTACT_CATEGORY_LIST);
 
-	changed = value_changed (cur_categories, new_categories);
-	g_free (new_categories);
-	g_free (cur_categories);
-
-	if (!changed)
+	/* Dont send a NULL array to the server */
+	if (!categories_list)
 		return;
 
+	prop_array = g_ptr_array_new ();
 
-#ifdef DEBUG
-	printf ("CATEGORIES = %s\n", new_categories);
-#endif
-
-	g_object_get ((new_contact->card),
-			"category_list", &categories,
-			NULL);
-
-	iter = e_list_get_iterator (categories);
-	while (e_iterator_is_valid (iter)) {
-		const char *category = e_iterator_get (iter);
-
-		if (!array)
-			array = g_ptr_array_new ();
-		g_ptr_array_add (array, g_strdup (category));
-		e_iterator_next (iter);
+	for (l = categories_list; l; l = g_list_next (l)) {
+		g_ptr_array_add (prop_array, g_strdup (l->data));
 	}
-	g_object_unref (iter);
 
-	e2k_properties_set_string_array (props, prop_mapping->prop_name, array);
+	e2k_properties_set_string_array (props, prop_mapping->prop_name, prop_array);
 }
 
 static void
-populate_categories(EContactField field, EContact *new_contact, void *data)
+populate_categories (EContactField field, EContact *new_contact, void *data)
 {
-	GSList *list = data;
-	GSList *l;
-	EList *categories;
+	GList *updated_list = NULL;
+	GPtrArray *categories_list = data;
+	gint i;
 
-	categories = e_list_new ((EListCopyFunc) g_strdup, 
-				 (EListFreeFunc) g_free,
-				 NULL);
-
-	for (l = list; l; l = l->next) {
-		e_list_append (categories, l->data);
+	for (i = 0; i < categories_list->len; i++) {
+		updated_list = g_list_append (updated_list,
+					      (char *)categories_list->pdata[i]);
 	}
+	
 
-	g_object_set ((new_contact->card),
-			"category_list", categories,
-			NULL);
+	e_contact_set (new_contact, E_CONTACT_CATEGORY_LIST, updated_list);
 
-#ifdef DEBUG 
-	{
-		char *category_text = e_contact_get (new_contact, E_CONTACT_CATEGORIES);
-		printf ("populated categories = %s\n", category_text);
-		g_free (category_text);
-	}
-#endif
-
-	g_object_unref (categories);
 }
-#endif
 
 static void
 proppatch_date (PropMapping *prop_mapping,
@@ -997,6 +973,44 @@ proppatch_address (PropMapping *prop_mapping,
 		e_contact_address_free (cur_address);
 }
 
+static void
+proppatch_im (PropMapping *prop_mapping, 
+	      EContact *new_contact, EContact *cur_contact,
+	      E2kProperties *props)
+{
+	GList *new_list;
+
+	new_list = e_contact_get (new_contact, prop_mapping->field);
+	while (new_list) {
+		if (prop_mapping->field == E_CONTACT_IM_MSN) {
+			e2k_properties_set_string (props, 
+						   E2K_PR_OUTLOOK_CONTACT_IM_ADDR,
+						   g_strdup (new_list->data));
+			break;
+		}
+		new_list = g_list_next (new_list);
+	}
+}
+
+static void
+populate_im (EContactField field, EContact *new_contact, void *data)
+{
+	GList *im_attr_list = NULL;
+	EVCardAttribute *attr;
+
+	if (field == E_CONTACT_IM_MSN) {
+		attr = e_vcard_attribute_new ("", e_contact_vcard_attribute (field));
+		e_vcard_attribute_add_param_with_value (attr,
+				e_vcard_attribute_param_new (EVC_TYPE), "WORK");
+
+		e_vcard_attribute_add_value (attr, (const char *)data);
+
+		im_attr_list = g_list_append (im_attr_list, attr);
+	}
+
+	e_contact_set_attributes (new_contact, field, im_attr_list);
+}
+
 static void
 populate_address (EContactField field, EContact *new_contact, void *data)
 {
@@ -1119,7 +1133,7 @@ static GByteArray *
 build_message (const char *from_name, const char *from_email,
 	       const char *subject, const char *note, EContactPhoto *photo)
 {
-	CamelDataWrapper *wrapper;
+	CamelDataWrapper *wrapper = NULL;
 	CamelContentType *type;
 	CamelMimeMessage *msg;
 	CamelInternetAddress *from;
@@ -1274,10 +1288,13 @@ e_book_backend_exchange_create_contact (EBookBackendSync  *backend,
 
 	d(printf("ebbe_create_contact(%p, %p, %s)\n", backend, book, vcard));
 
+	g_mutex_lock (bepriv->create_mutex);
+
 	switch (bepriv->mode) {
 
 	case GNOME_Evolution_Addressbook_MODE_LOCAL:
 		*contact = NULL;
+		g_mutex_unlock (bepriv->create_mutex);
 		return GNOME_Evolution_Addressbook_RepositoryOffline;
 	
 	case GNOME_Evolution_Addressbook_MODE_REMOTE:	
@@ -1317,15 +1334,18 @@ e_book_backend_exchange_create_contact (EBookBackendSync  *backend,
 			e_book_backend_summary_add_contact (bepriv->summary,
 							    *contact);
 			e_book_backend_cache_add_contact (bepriv->cache, *contact);
+			g_mutex_unlock (bepriv->create_mutex);
 			return GNOME_Evolution_Addressbook_Success;
 		} else {
 			g_object_unref (*contact);
 			*contact = NULL;
+			g_mutex_unlock (bepriv->create_mutex);
 			return http_status_to_pas (status);
 		}
 	default:
 		break;
 	}
+	g_mutex_unlock (bepriv->create_mutex);
 	return GNOME_Evolution_Addressbook_Success;
 }
 
@@ -2081,8 +2101,8 @@ e_book_backend_exchange_get_contact (EBookBackendSync  *backend,
 }
 
 
-static EBookBackendSyncStatus
-e_book_backend_exchange_authenticate_user (EBookBackendSync *backend,
+static void
+e_book_backend_exchange_authenticate_user (EBookBackend *backend,
 					   EDataBook        *book,
 					   guint32 	     opid,
 					   const char       *user,
@@ -2091,6 +2111,8 @@ e_book_backend_exchange_authenticate_user (EBookBackendSync *backend,
 {
 	EBookBackendExchange *be = E_BOOK_BACKEND_EXCHANGE (backend);
 	EBookBackendExchangePrivate *bepriv = be->priv;
+	ExchangeAccountResult result;
+	ExchangeAccount *account;
 
 	d(printf("ebbe_authenticate_user(%p, %p, %s, %s, %s)\n", backend, book, user, password, auth_method));
 
@@ -2100,10 +2122,20 @@ e_book_backend_exchange_authenticate_user (EBookBackendSync *backend,
 		e_book_backend_notify_writable (E_BOOK_BACKEND (backend), FALSE);
 		e_book_backend_notify_connection_status (E_BOOK_BACKEND (backend), FALSE);
 		e_data_book_respond_authenticate_user (book, opid, GNOME_Evolution_Addressbook_Success);
-		return GNOME_Evolution_Addressbook_Success;
+		return;
 			
 	case GNOME_Evolution_Addressbook_MODE_REMOTE:
-
+		
+		account = exchange_component_get_account_for_uri (global_exchange_component, NULL);
+		/* FIXME : Check for failures */
+		if (!exchange_account_get_context (account)) {
+			if(!exchange_account_connect (account, password, &result)) {
+				e_data_book_respond_authenticate_user (book, opid, GNOME_Evolution_Addressbook_AuthenticationFailed);
+				return;
+			}
+		}
+		if (!bepriv->connected);
+			e_book_backend_exchange_connect (be);
 		if (e_book_backend_cache_is_populated (bepriv->cache)) {
 			if (bepriv->is_writable)
 				g_thread_create ((GThreadFunc) update_cache, 
@@ -2113,14 +2145,35 @@ e_book_backend_exchange_authenticate_user (EBookBackendSync *backend,
 			/* for personal books we always cache*/
 			g_thread_create ((GThreadFunc) build_cache, be, FALSE, NULL);
 		}
-		return GNOME_Evolution_Addressbook_Success;
+		e_data_book_respond_authenticate_user (book, opid, GNOME_Evolution_Addressbook_Success);
+		return;
 
 	default:
 		break;
 	}
-	return GNOME_Evolution_Addressbook_Success;
+	e_data_book_respond_authenticate_user (book, opid, GNOME_Evolution_Addressbook_Success);
+	return;
 }
 
+static void
+e_book_backend_exchange_get_supported_auth_methods (EBookBackend *backend,
+						    EDataBook *book,
+						    guint32 opid)
+{
+	GList *auth_methods = NULL;
+	char *auth_method;
+
+	d(printf ("ebbe_get_supported_auth_methods (%p, %p)\n", backend, book));
+
+	auth_method = g_strdup_printf ("plain/password");
+	auth_methods = g_list_append (auth_methods, auth_method);
+	e_data_book_respond_get_supported_auth_methods (book, opid,
+				GNOME_Evolution_Addressbook_Success,
+				auth_methods);
+
+	g_free (auth_method);
+	g_list_free (auth_methods);
+}
 
 static EBookBackendSyncStatus
 e_book_backend_exchange_get_supported_fields (EBookBackendSync  *backend,
@@ -2182,7 +2235,6 @@ e_book_backend_exchange_load_source (EBookBackend *backend,
 {
 	EBookBackendExchange *be = E_BOOK_BACKEND_EXCHANGE (backend);
 	EBookBackendExchangePrivate *bepriv = be->priv;
-	GNOME_Evolution_Addressbook_CallStatus status;
 	const char *offline;
 
 	g_return_val_if_fail (bepriv->connected == FALSE, GNOME_Evolution_Addressbook_OtherError);
@@ -2219,9 +2271,7 @@ e_book_backend_exchange_load_source (EBookBackend *backend,
 		return GNOME_Evolution_Addressbook_Success;
 	}
 
-	status = e_book_backend_exchange_connect (be);
-	if (status != GNOME_Evolution_Addressbook_Success)
-		return status;
+	e_book_backend_set_is_loaded (E_BOOK_BACKEND (be), TRUE);
 
 	if (e_book_backend_cache_is_populated (bepriv->cache)) {
 		if (bepriv->is_writable)
@@ -2233,15 +2283,21 @@ e_book_backend_exchange_load_source (EBookBackend *backend,
 		g_thread_create ((GThreadFunc) build_cache, be, FALSE, NULL);
 	}
 
-	return status;
+	return GNOME_Evolution_Addressbook_Success;
 }
 
 static EBookBackendSyncStatus
 e_book_backend_exchange_remove (EBookBackendSync *backend, EDataBook *book, guint32 opid)
 {
-	d(printf("ebbe_remove(%p, %p)\n", backend, book));
-	return GNOME_Evolution_Addressbook_PermissionDenied;
+	EBookBackendExchange *be = E_BOOK_BACKEND_EXCHANGE (backend);
+	EBookBackendExchangePrivate *bepriv = be->priv;
+	ExchangeAccountFolderResult result;
+	const char *uri;
 
+	d(printf("ebbe_remove(%p, %p)\n", backend, book));
+	uri = e_folder_exchange_get_internal_uri (bepriv->folder);
+	result = exchange_account_remove_folder (bepriv->account, uri);
+	return GNOME_Evolution_Addressbook_Success;
 	/* FIXME: Folder deletion from contacts view */
 #if 0 
 	EBookBackendExchange *be = E_BOOK_BACKEND_EXCHANGE (backend);
@@ -2288,9 +2344,10 @@ e_book_backend_exchange_set_mode (EBookBackend *backend, int mode)
 {
 	EBookBackendExchange *be = E_BOOK_BACKEND_EXCHANGE (backend);
 	EBookBackendExchangePrivate *bepriv = be->priv;
+	ExchangeAccount *account;
 
 	bepriv->mode = mode;
-	if (e_book_backend_is_loaded (backend)) {
+	/* if (e_book_backend_is_loaded (backend)) { */
 		if (mode == GNOME_Evolution_Addressbook_MODE_LOCAL) {
 			e_book_backend_set_is_writable (backend, FALSE);
 			e_book_backend_notify_writable (backend, FALSE);
@@ -2300,10 +2357,11 @@ e_book_backend_exchange_set_mode (EBookBackend *backend, int mode)
 			e_book_backend_set_is_writable (backend, bepriv->is_writable);
 			e_book_backend_notify_writable (backend, bepriv->is_writable);
 			e_book_backend_notify_connection_status (backend, TRUE);
-			/* FIXME :
-			e_book_backend_notify_auth_required (backend); */
+			account = exchange_component_get_account_for_uri (global_exchange_component, NULL);
+			if (!exchange_account_get_context (account))
+				e_book_backend_notify_auth_required (backend);
 		}
-	}
+/*	} */
 }
 
 /**
@@ -2357,6 +2415,10 @@ e_book_backend_exchange_dispose (GObject *object)
 		if (be->priv->cache)
 			g_object_unref (be->priv->cache);
 
+		if (be->priv->create_mutex)
+			g_mutex_free (be->priv->create_mutex);
+			
+
 		g_free (be->priv);
 		be->priv = NULL;
 	}
@@ -2395,6 +2457,9 @@ e_book_backend_exchange_class_init (EBookBackendExchangeClass *klass)
 	backend_class->stop_book_view          = e_book_backend_exchange_stop_book_view;
 	backend_class->cancel_operation        = e_book_backend_exchange_cancel_operation;
 	backend_class->set_mode			= e_book_backend_exchange_set_mode;
+	backend_class->get_supported_auth_methods = e_book_backend_exchange_get_supported_auth_methods;
+	backend_class->authenticate_user     = e_book_backend_exchange_authenticate_user;
+
 	sync_class->remove_sync                = e_book_backend_exchange_remove;
 	sync_class->create_contact_sync        = e_book_backend_exchange_create_contact;
 	sync_class->remove_contacts_sync       = e_book_backend_exchange_remove_contacts;
@@ -2402,7 +2467,6 @@ e_book_backend_exchange_class_init (EBookBackendExchangeClass *klass)
 	sync_class->get_contact_sync           = e_book_backend_exchange_get_contact;
 	sync_class->get_contact_list_sync      = e_book_backend_exchange_get_contact_list;
 	sync_class->get_changes_sync           = e_book_backend_exchange_get_changes;
-	sync_class->authenticate_user_sync     = e_book_backend_exchange_authenticate_user;
 	sync_class->get_supported_fields_sync  = e_book_backend_exchange_get_supported_fields;
 	sync_class->get_required_fields_sync   = e_book_backend_exchange_get_required_fields;
 
@@ -2422,6 +2486,8 @@ e_book_backend_exchange_init (EBookBackendExchange *backend)
 	priv->original_uri 	= NULL;
 	priv->is_writable 	= TRUE;
 
+	priv->create_mutex      = g_mutex_new ();
+
 	backend->priv 		= priv;
 }
 
